#!/bin/python3
#%%
import base64
import datetime
import functools
import gzip
import json
import os
from itertools import product
from random import shuffle
import re
from time import time, sleep

import numpy as np
import pandas as pd

from sklearn.linear_model import LogisticRegression
from sklearn.metrics import log_loss
from sklearn.model_selection import train_test_split
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.ensemble import HistGradientBoostingClassifier
from xgboost import XGBClassifier
from lightgbm import LGBMClassifier

# from flaml import AutoML
from hyperopt import hp, fmin, tpe, Trials, STATUS_OK, space_eval

from sklearn2pmml import sklearn2pmml
from sklearn2pmml.pipeline import PMMLPipeline
from pypmml import Model

from tools.data_processing.basic_data_tools import pickle_load
from tools.std_lib.basic import json_dumps, json_dump, json_load
from tools.data_processing.easy_pandas import dict_to_json, get_bytes_sha256, get_file_hash, get_timestamp, ls_dirs, ls_dirs_r, multiprocessing_apply, pickle_dump, read_txt, save_and_reset_index, t_print, get_hash, date_str, get_uuid_hex, save_txt

#%%

def get_cut_points(s: pd.Series, bins = 20, endpoints = [-float('inf'), float('inf')]):
    points = []
    s = pd.Series(s).copy()
    s.fillna(s.mean(), inplace=True)
    s.sort_values(inplace=True)
    for k in range(1, bins):
        n = int(len(s) * k / bins)
        cut = s.iloc[n]
        points.append(cut)
    n1 = len(points)
    points = sorted(set(points + endpoints))
    n2 = len(points)
    print(f'get_cut_points: info: bins = {n2-1}')
    if n1 + len(endpoints) != n2:
        print('get_cut_points: warning: too many duplicated numbers.')
    points = sorted(points)
    return points


def get_gain_and_lift(y, y_proba, bins = 'auto', prefix = '', y_proba_max = 1):
    if bins == 'auto':
        bins = get_cut_points(y_proba, bins=20, endpoints=[0, y_proba_max])
    bins = sorted(bins, reverse=True)
    bins = [k for k in bins if k < y_proba_max]
    print('bins =', bins)
    df = pd.DataFrame(
        index=range(1, len(bins)), 
        columns=[
            'interval_left', 'interval_right', 
            'total_count', 'pos_count', 'total_pct', 'pos_ratio', 
            'gain', 'lift',
            ],
        dtype=int)
    y_mean = y.mean()
    y_sum = y.sum()
    for k, cut in enumerate(bins):
        k = k+1
        idx = y_proba > cut
        cut_y = y[idx]
        cnt = len(cut_y)
        cnt_rate = cnt / len(y)
        pos_cnt = cut_y.sum()
        pos_rate = pos_cnt / cnt
        lift = pos_cnt / cnt / y_mean
        # df.loc[k, 'cut_point'] = cut
        df.loc[k, 'interval_left'] = cut
        df.loc[k, 'interval_right'] = y_proba_max
        df.loc[k, 'total_count'] = cnt
        df.loc[k, 'total_pct'] = cnt_rate
        df.loc[k, 'gain'] = pos_cnt / y_sum
        df.loc[k, 'pos_ratio'] = pos_rate
        df.loc[k, 'pos_count'] = pos_cnt
        df.loc[k, 'lift'] = lift
    df = df.loc[df['pos_count'] > 0, :]
    df = df.add_prefix(prefix)
    df.insert(0, 'interval', df.apply(lambda row: f"({row['interval_left']:.9f}, {row['interval_right']:.9f}]", axis=1))
    df.insert(0, 'subset', (((df['total_pct']*100).round()/5).round()*5).apply(lambda s: f'top{s:0.0f}'))
    df['total_count'] = df['total_count'].astype(int)
    df['pos_count'] = df['pos_count'].astype(int)
    df.index.name = 'index'
    print('lift =')
    print(df)
    return df


def get_tri_lift(Y3, Y3_proba):
    y_train, y_valid, y_test = Y3
    y_proba_train, y_proba_valid, y_proba_test = Y3_proba
    train_df = get_gain_and_lift(y_train, y_proba_train)
    valid_df = get_gain_and_lift(y_valid, y_proba_valid)
    test_df = get_gain_and_lift(y_test, y_proba_test)
    tri_lift_dict = {
        'train': train_df,
        'validate': valid_df,
        'test': test_df
    }
    return tri_lift_dict


def get_distribution_by_interval(y_proba: pd.Series, bins = 'auto', y_proba_max = 1):
    y_proba = pd.Series(y_proba)
    if bins == 'auto':
        bins = get_cut_points(y_proba, bins=20, endpoints=[0, y_proba_max])
    def func(s):
        sr = pd.cut(s, bins=bins, precision=9).value_counts().sort_index()
        sr['Not_a_Number'] = s.isna().sum()
        return sr
    df = pd.DataFrame({
        'cnt': func(y_proba),
    })
    df['pct'] = df['cnt'] / df['cnt'].sum()
    df.index.name = 'interval'
    df = save_and_reset_index(df)
    df['interval_left'] = df['interval'].apply(lambda s: s.left if isinstance(s, pd.Interval) else float('nan'))
    df['interval_right'] = df['interval'].apply(lambda s: s.right if isinstance(s, pd.Interval) else float('nan'))
    cols = ['interval', 'interval_left', 'interval_right', 'cnt', 'pct']
    df = df[cols]
    return df


def get_tri_distribution_by_interval(Y3, Y3_proba):
    y_train, y_valid, y_test = Y3
    y_proba_train, y_proba_valid, y_proba_test = Y3_proba
    d = {
        'y_proba_train': get_distribution_by_interval(y_proba_train),
        'y_proba_validate': get_distribution_by_interval(y_proba_valid),
        'y_proba_test': get_distribution_by_interval(y_proba_test),
        
        'y_proba_train_1': get_distribution_by_interval(y_proba_train[y_train==1]),
        'y_proba_validate_1': get_distribution_by_interval(y_proba_valid[y_valid==1]),
        'y_proba_test_1': get_distribution_by_interval(y_proba_test[y_test==1]),
        
        'y_proba_train_0': get_distribution_by_interval(y_proba_train[y_train==0]),
        'y_proba_validate_0': get_distribution_by_interval(y_proba_valid[y_valid==0]),
        'y_proba_test_0': get_distribution_by_interval(y_proba_test[y_test==0]),
    }
    return d


def get_cdf_curve(s: pd.Series):
    if not isinstance(s, pd.Series):
        s = pd.Series(s)
    s = s.copy()
    s.sort_values(inplace=True)
    n = len(s)
    df = pd.DataFrame([{
        'cnt': round(k/99*(n-1)),
        'pct': k/100, 
        'value': s.iloc[round(k/99*(n-1))]
        } for k in range(100)])
    df.index.name = 'index'
    return df


def get_tri_cdf_curve(Y3, Y3_proba):
    y_train, y_valid, y_test = Y3
    y_proba_train, y_proba_valid, y_proba_test = Y3_proba
    d = {
        'y_proba_train': get_cdf_curve(y_proba_train),
        'y_proba_validate': get_cdf_curve(y_proba_valid),
        'y_proba_test': get_cdf_curve(y_proba_test),
        
        'y_proba_train_1': get_cdf_curve(y_proba_train[y_train==1]),
        'y_proba_validate_1': get_cdf_curve(y_proba_valid[y_valid==1]),
        'y_proba_test_1': get_cdf_curve(y_proba_test[y_test==1]),
        
        'y_proba_train_0': get_cdf_curve(y_proba_train[y_train==0]),
        'y_proba_validate_0': get_cdf_curve(y_proba_valid[y_valid==0]),
        'y_proba_test_0': get_cdf_curve(y_proba_test[y_test==0]),
    }
    return d


def simple_describe(s: np.array):
    s = pd.Series(s)
    d = s.describe().to_dict()
    d['count'] = int(d['count'])
    d['size'] = len(s)
    return d


def get_tri_describe(Y3, Y3_proba):
    y_train, y_valid, y_test = Y3
    y_proba_train, y_proba_valid, y_proba_test = Y3_proba
    d = {
        'y_proba_train': simple_describe(y_proba_train),
        'y_proba_validate': simple_describe(y_proba_valid),
        'y_proba_test': simple_describe(y_proba_test),
        
        'y_proba_train_1': simple_describe(y_proba_train[y_train==1]),
        'y_proba_validate_1': simple_describe(y_proba_valid[y_valid==1]),
        'y_proba_test_1': simple_describe(y_proba_test[y_test==1]),
        
        'y_proba_train_0': simple_describe(y_proba_train[y_train==0]),
        'y_proba_validate_0': simple_describe(y_proba_valid[y_valid==0]),
        'y_proba_test_0': simple_describe(y_proba_test[y_test==0]),
    }
    return d


def get_FI_df(features: list, clf: object):
    if isinstance(clf, LGBMClassifier):
        clf.importance_type = 'split'
        split_dict = dict(zip(features, clf.feature_importances_))
        clf.importance_type = 'gain'
        gain_dict = dict(zip(features, clf.feature_importances_))
        FI_df = pd.DataFrame({
            'split_origin': split_dict,
            'gain_origin': gain_dict,
        })
        FI_df['split_normalized'] = FI_df['split_origin'] / FI_df['split_origin'].sum()
        FI_df['gain_normalized'] = FI_df['gain_origin'] / FI_df['gain_origin'].sum()
        FI_df.index.name = 'feature'
        t_print('FI_df =')
        t_print(FI_df)
        FI_df = save_and_reset_index(FI_df)
    elif isinstance(clf, XGBClassifier):
        FI_dict = {}
        importance_type_ls = ['weight', 'gain', 'cover', 'total_gain', 'total_cover']
        for importance_type in importance_type_ls:
            clf.importance_type = importance_type
            FI_dict[importance_type] = clf.feature_importances_
        FI_df = pd.DataFrame(FI_dict, index=features)
        FI_df.index.name = 'feature'
        t_print('FI_df =')
        t_print(FI_df)
        FI_df = save_and_reset_index(FI_df)
        clf.importance_type = None
    return FI_df


def normalize_df(df: pd.DataFrame, verbose = False):
    norm_df = (df - df.mean()) / df.std()
    if verbose:
        t_print('describe =', norm_df.describe().T)
    return norm_df


def get_LR_coef(clf, X_df, normalized = False, to_series = True):
    coef = dict(zip(X_df.columns.to_list(), clf.coef_[0]))
    coef = pd.Series(coef)
    if normalized:
        coef = coef.abs()
        coef = coef / coef.sum()
    return coef


def get_LR_FI(clf, X_train, y_train):
    X_norm_train = normalize_df(X_train)
    norm_clf = LogisticRegression(max_iter = clf.max_iter, random_state = 0)
    norm_clf.fit(X_norm_train, y_train)
    describe_df = X_train.describe().T
    describe_df['count'] = describe_df['count'].astype(int)
    FI_df = pd.DataFrame({
        'origin_coef': get_LR_coef(clf, X_train),
        'norm_coef': get_LR_coef(norm_clf, X_norm_train),
        'feature_importances': get_LR_coef(norm_clf, X_train, normalized = True),
    })
    stat_df = pd.concat([describe_df, FI_df], axis = 1)
    save_and_reset_index(stat_df, 'feature')
    return stat_df


def get_feature_info(clf, X3: list, Y3: list):
    features = X3[0].columns.to_list()
    if isinstance(clf, LGBMClassifier):
        FI_df = get_FI_df(features=features, clf=clf)
        s_idx = FI_df['split_origin'] != 0
        g_idx = FI_df['gain_origin'] != 0
        if s_idx.equals(g_idx):
            idx = s_idx
            features_in_use = FI_df['feature'][idx].to_list()
            features_in_use_cnt = len(features_in_use)
        d = {'LGBMClassifier': {
            'feature_importances_stat': FI_df,
            'features_in_use': features_in_use,
            'features_in_use_cnt': features_in_use_cnt,}}
    elif isinstance(clf, XGBClassifier):
        FI_df = get_FI_df(features=features, clf=clf)
        idx = FI_df['gain'] != 0
        features_in_use = FI_df['feature'][idx].to_list()
        features_in_use_cnt = len(features_in_use)
        d = {'XGBClassifier': {
            'feature_importances_stat': FI_df,
            'features_in_use': features_in_use,
            'features_in_use_cnt': features_in_use_cnt,}
            }
    elif isinstance(clf, LogisticRegression):
        X_train = X3[0]
        y_train = Y3[0]
        FI_stat_df = get_LR_FI(clf, X_train, y_train)
        d = {'LogisticRegression': {
            'feature_importances_stat': FI_stat_df,
        }}
        t_print(FI_stat_df)
    return d


def more_func(clf, X3: list, Y3: list):
    if 'model' in dir(clf):
        clf = clf.model.estimator
    Y3_proba = [
        pd.Series(clf.predict_proba(xs)[:, 1], index=ys.index)
        for xs, ys in zip(X3, Y3)
        ]
    ret = {}
    ret['evaluation'] = {}
    ret['evaluation']['gain_and_lift'] = get_tri_lift(Y3, Y3_proba)
    ret['evaluation']['distribution_by_interval'] = get_tri_distribution_by_interval(Y3, Y3_proba)
    ret['evaluation']['cdf_curve'] = get_tri_cdf_curve(Y3, Y3_proba)
    ret['evaluation']['describe'] = get_tri_describe(Y3, Y3_proba)
    ret['feature'] = {}
    ret['feature']['feature_importances'] = get_feature_info(clf, X3, Y3)
    return ret


def get_params_dict_list(all_params_dict):

    all_keys = list(all_params_dict.keys())
    for k in all_keys:
        if all_params_dict[k] == []:
            del all_params_dict[k]

    keys = all_params_dict.keys()
    values = [k if type(k) == list else [k] for k in all_params_dict.values()]
    params_dict_list = [dict(zip(keys, combination)) for combination in product(*values)]
    
    r = []
    for k in params_dict_list:
        if k not in r:
            if k['clf_tag'] == 'AutoML':
                if k['AutoML.metric_alpha'] * k['AutoML.cv_score_agg_func_alpha'] != 0:
                    continue
            r.append(k)

    return r


def get_sha256(s, n = 0):
    if type(s) == bytes:
        r = get_bytes_sha256(s)
    elif type(s) == str:
        s = s.encode('utf-8')
        r = get_sha256(s)
    elif type(s) == list or type(s) == dict:
        s = json_dumps(dict_to_json(s))
        r = get_sha256(s)
    elif type(s) == pd.core.indexes.numeric.Int64Index:
        s = s.to_list()
        r = get_sha256(s)
    if n > 0:
        r = r[:n]
    return r


short_param = {
    # 'random_state': 'rs',

    # GradientBoostingClassifier
    'learning_rate': 'lr', 
    'n_estimators': 'ne', 
    'subsample': 'ss', 
    'min_samples_split': 'ms', 
    'min_samples_leaf': 'ml', 
    'max_depth': 'md', 
    'max_leaf_nodes': 'mn', 
    
    # HistGradientBoostingClassifier
    'learning_rate': 'lr', 
    'max_iter': 'mi', 
    'max_leaf_nodes': 'mn', 
    'max_depth': 'md', 
    'min_samples_leaf': 'ml', 
    'l2_regularization': 'lt', 

    # LGBMClassifier
    'num_leaves': 'nl', 
    'max_depth': 'md', 
    'learning_rate': 'lr', 
    'n_estimators': 'ne', 
    'min_split_gain': 'mg', 
    'min_child_weight': 'mw', 
    'min_child_samples': 'ml', 
    'subsample': 'ss', 
    'subsample_freq': 'sf', 
    'colsample_bytree': 'ct', 
    'reg_alpha': 'ra', 
    'reg_lambda': 'rl', 
    'class_weight': 'cw',
    'n_jobs': 'nj', 

    # XGBClassifier
    'n_estimators': 'ne', 
    'max_depth': 'md', 
    'max_leaves': 'nl', 
    'learning_rate': 'lr', 
    'n_jobs': 'nj', 
    'min_child_weight': 'mw', 
    'subsample': 'ss', 
    'colsample_bytree': 'ct', 
    'colsample_bylevel': 'cl', 
    'colsample_bynode': 'cn', 
    'reg_alpha': 'ra', 
    'reg_lambda': 'rl', 
    
    # LR
    'penalty': 'pn',
    'class_weight': 'cw',
    'max_iter': 'mi',
    'l1_ratio': 'lo',
    'solver': 'sv',
    'n_jobs': 'nj', 
    'tol': 'tl', 

    # AutoML (possible?!)
    }


LR_solver_dict = {
    'lbfgs': '0',
    'liblinear': '1',
    'newton-cg': '2',
    'newton-cholesky': '3',
    'sag': '4',
    'saga': '5'
    }


def short_num(a):
    if type(a) == str:
        if a == 'balanced':
            r = '1'
        elif a == 'elasticnet':
            r = '12'
        elif a == 'l1':
            r = '1'
        elif a == 'l2':
            r = '2'
        elif a in LR_solver_dict:
            r = LR_solver_dict[a]
    elif type(a) == int or abs(a - round(a)) < 1e-6:
        a = round(a)
        r = str(a)
    elif type(a) == float and a < 1:
        r = str(a).replace('.', '').rstrip('0')
    else:
        print('warning: short_num: stupid type?!', a)
    return r


def params_dict_to_params_str(params):
    d = params.copy()

    for del_key in ['random_state', 'n_jobs']:
        if del_key in d:
            del d[del_key]

    if len(d) > 0:
        s = ''.join([
            short_param[k] + short_num(v)
            for k, v in sorted(d.items())
            if v != None
        ])
    else:
        s = 'default'

    return s


def fix_settings(settings, X):

    origin_setting_sha256 = get_sha256(settings, 8)

    if type(X) == list:
        X_index_ls = [Xk.index.to_list() for Xk in X]
        X = X[0]
    else:
        X_index_ls = X.index.to_list()
    
    cols = X.columns.to_list()
    if 'dim_name_func' in settings:
        dim_name_func = settings['dim_name_func']
        # dim_str = get_com_dim_name(settings['selected_dim'])
        # cols_sha256 = get_sha256(settings['selected_dim'], 8)
        # dim_str = f'dim_{cols_sha256}'
        if 'selected_dim' in settings:
            selected_dim = settings['selected_dim']
        else:
            selected_dim = cols
        dim_str = dim_name_func(selected_dim)
    elif 'selected_dim' in settings:
        cols_sha256 = get_sha256(settings['selected_dim'], 8)
        dim_str = f'dim_{cols_sha256}'
    else:
        # idxs_sha256 = get_sha256(X_index_ls, 4)
        # cols_sha256 = get_sha256(cols, 4)
        # dim_str = f'dim_{idxs_sha256}_{cols_sha256}'
        cols_sha256 = get_sha256(cols, 8)
        dim_str = f'dim_{cols_sha256}'

    if 'clf_tag' not in settings:
        settings['clf_tag'] = 'GB'

    clf_tag = settings['clf_tag']

    # if clf_tag == 'AutoML':
    #     t_print('warning: AutoML log will not be saved.')
    #     if 'exp_name' not in settings:
    #         exp_name = 'AutoML_' + get_sha256(settings, 8)
    #         settings['exp_name'] = exp_name
    #         t_print(f'warning: exp_name was missing and let expname = {exp_name}.')

    params_keys = [k for k in settings if k.startswith(clf_tag)]
    params = {k[len(clf_tag)+1:]: settings[k] for k in params_keys}
    for k in params_keys:
        del settings[k]

    if 'DS.random_state' not in settings:
        settings['DS.random_state'] = 0
    ds_str = '-ds' + str(settings['DS.random_state'])

    if 'DS.test_size' not in settings:
        settings['DS.test_size'] = .2
        
    ds_str = '-ds' + str(settings['DS.random_state'])


    if 'random_state' in params:
        rs = params['random_state']
    else:
        rs = 0

    if 'exp_name' not in settings:
        prefix = settings['prefix']
        clf_tag = settings['clf_tag']
        if clf_tag in ['AutoML', 'hyperopt']:
            params_str = get_hash(params, n = 8)
        else:
            params_str = params_dict_to_params_str(params)
        exp_name = f'{prefix}-{clf_tag}_{params_str}-{dim_str}'
        settings['exp_name'] = exp_name
    elif not settings['exp_name'].startswith(settings['prefix']):
        settings['exp_name'] = f"{settings['prefix']}-{settings['exp_name']}"
    exp_name = settings['exp_name']

    save_name_0 = f'{exp_name}_rs{rs}{ds_str}-'
    settings['save_name_0'] = save_name_0

    return settings, params


replace_slash = lambda p: p.replace('\\', '/')


def run_clf(settings):

    r = True

    results_path = settings['results_path']
    prefix = settings['prefix']
    exp_name = settings['exp_name']
    save_name_0 = settings['save_name_0']

    exp_path = os.path.join(results_path, prefix, exp_name)

    if os.path.exists(exp_path):
        exps = [k[:-17] for k in os.listdir(exp_path)]
        if save_name_0 in exps:
            r = False
            save_name_0_path = os.path.join(exp_path, save_name_0)
            save_name_0_path = replace_slash(save_name_0_path)
            t_print(f'{save_name_0_path}* exists')

    return r


def get_xy_tri(X, y, ds = 0, test_size = .2):
    if len(X) == len(y) == 2:
        X_sample, X_test = X
        y_sample, y_test = y
        X_train, X_valid, y_train, y_valid = train_test_split(X_sample, y_sample, test_size=test_size, random_state=ds, stratify=y_sample)
    elif len(X) == len(y) == 3:
        X_train, X_valid, X_test = X
        y_train, y_valid, y_test = y
        if ds != 0:
            del test_size
            test_size = len(y_valid) / (len(y_train) + len(y_valid))
            t_print('(estimated) test_size =', test_size, ', random_stat =', ds)
            X_sample = pd.concat([X_train, X_valid]).sort_index()
            y_sample = pd.concat([y_train, y_valid]).sort_index()
            X_train, X_valid, y_train, y_valid = train_test_split(X_sample, y_sample, test_size=test_size, random_state=ds, stratify=y_sample)
    elif len(X) == len(y) > 3:
        X_sample, X_test, y_sample, y_test = train_test_split(X, y, test_size=test_size, random_state=0, stratify=y)
        X_train, X_valid, y_train, y_valid = train_test_split(X_sample, y_sample, test_size=test_size, random_state=ds, stratify=y_sample)
    return X_train, X_valid, X_test, y_train, y_valid, y_test


def custom_metric_alpha(X_val, y_val, estimator, labels, X_train, y_train,
                  weight_val=None, weight_train=None, config=None,
                  groups_val=None, groups_train=None, alpha = .5):

    y_train_pred = estimator.predict_proba(X_train)
    train_loss = log_loss(y_train, y_train_pred, labels=labels, sample_weight=weight_train)
    
    y_val_pred = estimator.predict_proba(X_val)
    val_loss = log_loss(y_val, y_val_pred, labels=labels, sample_weight=weight_val)
    
    # alpha = .5
    minimized = val_loss + alpha*(val_loss - train_loss) 

    logged = {
        "train_loss": train_loss, 
        "val_loss": val_loss, 
        'train_auc': roc_auc_score(y_train, y_train_pred[:, 1]),
        'val_auc': roc_auc_score(y_val, y_val_pred[:, 1]),
        'train_ks': ks_score(y_train, y_train_pred[:, 1]),
        'val_ks': ks_score(y_val, y_val_pred[:, 1]),

        'train_len': len(y_train), 
        'val_len': len(y_val), 
        'train_pos_rat': sum(y_train) / len(y_train),
        'val_pos_rat': sum(y_val) / len(y_val),
    }
    
    return minimized, logged


def get_custom_metric(alpha = .5):
    return functools.partial(custom_metric_alpha, alpha = alpha) 

iter_global = 0

def cv_score_agg_func_alpha(val_loss_folds, log_metrics_folds, alpha = 0):
    metric_to_minimize = np.mean(val_loss_folds) + alpha*np.std(val_loss_folds)
    metrics_to_log = log_metrics_folds
    d = {
        'train_auc': [], 
        'val_auc': [], 
        'train_ks': [], 
        'val_ks': [], 
        }
    for f in log_metrics_folds:
        for m in d:
            d[m].append(f[m])
    rel_err = (np.mean(d['train_ks']-np.mean(d['val_ks']))) / np.mean(d['val_ks']) if np.mean(d['val_ks']) != 0 else float('nan')
    for k in d:
        d[k] = f'{np.mean(d[k]):.3f}+-{np.std(d[k]):.3f}' 
    global iter_global
    iter_global += 1
    d['rel_err'] = f'{rel_err:.3f}'
    d['iter'] = iter_global
    d['exp_name'] = exp_name_global
    t_print(d)
    return metric_to_minimize, metrics_to_log


def get_cv_score_agg_func(alpha = 0):
    return functools.partial(cv_score_agg_func_alpha, alpha = alpha) 


def get_clf(clf_tag: str, params: dict):

    params2 = params.copy()

    if 'metric_alpha' in params2:
        alpha = params2['metric_alpha']
        del params2['metric_alpha']
        params2['metric'] = get_custom_metric(alpha)

    if 'cv_score_agg_func_alpha' in params2:
        alpha = params2['cv_score_agg_func_alpha']
        del params2['cv_score_agg_func_alpha']
        params2['cv_score_agg_func'] = get_cv_score_agg_func(alpha)

    if clf_tag in ['XGB']:
        params2['verbosity'] = 1
    elif clf_tag in ['AutoML']:
        params2['verbose'] = 3
    else:
        params2['verbose'] = 1

    if clf_tag == 'GB':
        clf = GradientBoostingClassifier(**params2)
    elif clf_tag == 'HGB':
        clf = HistGradientBoostingClassifier(**params2)
    elif clf_tag == 'LGBM':
        clf = LGBMClassifier(**params2)
    elif clf_tag == 'XGB':
        clf = XGBClassifier(**params2)
    elif clf_tag == 'LR':
        clf = LogisticRegression(**params2)
    elif clf_tag == 'AutoML':
        clf = AutoML(**params2)
    else:
        t_print('No clf_tag found.')

    if 'random_state' in dir(clf) and clf.random_state == None:
        clf.random_state = 0
    elif 'random_state' not in dir(clf):
        t_print(f'warning: no random_state in {clf}')

    return clf


# hyperopt_func_v0_space = {
#     # "n_estimators": hp.choice("n_estimators", [20, 50, 100, 150, 200]),
#     # "n_estimators": hp.randint("n_estimators", 10, 101),
#     "n_estimators": hp.randint("n_estimators", 20, 101),
#     'learning_rate': hp.uniform('learning_rate', .01, .20),
#     # "max_depth": hp.choice("max_depth", [2, 3, 4, 5]),
#     "max_depth": hp.choice("max_depth", [1, 2, 3]),
#     "subsample": hp.uniform("subsample", 0.1, 1),
#     "colsample_bytree": hp.uniform("colsample_bytree", 0.1, 1),
#     "gamma": hp.uniform("gamma", .1, 10),
#     "reg_alpha": hp.uniform("reg_alpha", .1, 10),
#     # "min_child_weight": hp.choice('min_child_weight', [20, 50, 100, 200, 500]),
#     "min_child_weight": hp.randint("min_child_weight", 10, 1001),
#     }

hyperopt_func_v0_space = {
    "n_estimators": hp.randint("n_estimators", 50, 201),
    'learning_rate': hp.uniform('learning_rate', .001, .20),
    "max_depth": hp.choice("max_depth", [2, 3]),
    "subsample": hp.uniform("subsample", .5, 1),
    "colsample_bytree": hp.uniform("colsample_bytree", .5, 1),
    "gamma": hp.uniform("gamma", .1, 10),
    "reg_alpha": hp.uniform("reg_alpha", .1, 10),
    "min_child_weight": hp.randint("min_child_weight", 10, 1001),
    }

hyperopt_func_overfit_space = {
    "n_estimators": hp.randint("n_estimators", 20, 101),
    'learning_rate': hp.uniform('learning_rate', .001, .20),
    "max_depth": hp.choice("max_depth", [1, 2,]),
    "subsample": hp.uniform("subsample", .5, 1),
    "colsample_bytree": hp.uniform("colsample_bytree", .5, 1),
    "gamma": hp.uniform("gamma", .1, 10),
    "reg_alpha": hp.uniform("reg_alpha", .1, 10),
    "min_child_weight": hp.randint("min_child_weight", 10, 1001),
    }

hyperopt_func_overfit2_space = {
    "n_estimators": hp.randint("n_estimators", 20, 101),
    'learning_rate': hp.uniform('learning_rate', .001, .20),
    "max_depth": hp.choice("max_depth", [1, 2,]),
    "subsample": hp.uniform("subsample", .1, 1),
    "colsample_bytree": hp.uniform("colsample_bytree", .1, 1),
    "gamma": hp.uniform("gamma", 0, 100),
    "reg_alpha": hp.uniform("reg_alpha", 0, 100),
    "min_child_weight": hp.randint("min_child_weight", 10, 1001),
    }

hyperopt_func_space_dict = {
    'hyperopt_func_v0_space': hyperopt_func_v0_space,
    'hyperopt_func_overfit_space': hyperopt_func_overfit_space,
    'hyperopt_func_overfit2_space': hyperopt_func_overfit2_space,
}


def get_clf_auc_ks(clf, X, y, weight=None, cut_points = False):
    y_pred = clf.predict_proba(X)[:, 1]
    auc = roc_auc_score(y, y_pred, sample_weight=weight)
    ks = ks_score(y, y_pred, sample_weight=weight)
    ret = {'auc': auc, 'ks': ks, 'reverse': auc <= .5}
    if cut_points:
        ret['cut_points'] = get_cut_points(y_pred, endpoints=[0, 1])
    return ret


def to_numpy(var, fillna = None) -> np.array:
    if isinstance(var, pd.DataFrame) or isinstance(var, pd.Series) :
        # if fillna is not None:
        #     var = var.fillna(-10)
        var = var.to_numpy()
    elif isinstance(var, list):
        var = [to_numpy(k) for k in var]
    return var


def objective_func_with_params(params, XY3 = None, loss_func = None, fillna = None):

    if XY3 is not None: 
        XY3 = to_numpy(XY3, fillna=fillna)
        X3, Y3 = XY3
        X_train, X_validate, X_test = X3
        y_train, y_validate, y_test = Y3

    if True:
        params['scale_pos_weight'] = (y_train==0).sum() / (y_train==1).sum()

    t_print('model = XGBClassifier(**', params, ', random_state=0)')
    model = XGBClassifier(**params, random_state=0)
    t_print('model.fit(X_train, y_train)')
    model.fit(X_train, y_train)

    metric_dict = {
        'train': get_clf_auc_ks(clf=model, X=X_train, y=y_train, cut_points=True),
        'validate': get_clf_auc_ks(clf=model, X=X_validate, y=y_validate, cut_points=True),
        'test': get_clf_auc_ks(clf=model, X=X_test, y=y_test, cut_points=True),
    }

    for d in metric_dict.values():
        d['cut_points'] = len(d['cut_points']) - 1
    t_print(metric_dict)

    loss = loss_func(metric_dict)

    ret = {
        'loss': loss, 
        'status': STATUS_OK, 
        'metric_dict': metric_dict,
        }
    
    return ret


def loss_func_alpha(
        metric_dict: dict, 
        loss_alpha = 1, 
        loss_items = None, 
        reversed_as_inf = True, 
        min_bins = 0, 
        overfit_rate = None
        ):

    reverse = any([value['reverse'] for value in metric_dict.values()])
    bins_diff = np.mean([
        max([min_bins - value['cut_points'], 0]) 
        for value in metric_dict.values()
        ])

    validate_auc_loss = 2*(1 - metric_dict['validate']['auc'])
    test_auc_loss = 2*(1 - metric_dict['test']['auc'])

    validate_ks_loss = 1 - metric_dict['validate']['ks']
    test_ks_loss = 1 - metric_dict['test']['ks']

    validate_perf = (metric_dict['train']['auc']-.5) + metric_dict['train']['ks']/2
    test_perf = (metric_dict['test']['auc']-.5) + metric_dict['test']['ks']/2
    validate_loss = 1 - validate_perf
    test_loss = 1 - test_perf
    
    auc_overfit = abs(metric_dict['train']['auc'] - metric_dict['validate']['auc'])
    ks_overfit = abs(metric_dict['train']['ks'] - metric_dict['validate']['ks'])
    overfit = (2*auc_overfit + ks_overfit) / 2
    
    auc_decline = abs(metric_dict['validate']['auc'] - metric_dict['test']['auc'])
    ks_decline = abs(metric_dict['validate']['ks'] - metric_dict['test']['ks'])
    decline = (2*auc_decline + ks_decline) / 2

    loss_dict = {
        'validate_auc_loss': validate_auc_loss,
        'test_auc_loss': test_auc_loss,

        'validate_ks_loss': validate_ks_loss,
        'test_ks_loss': test_ks_loss,

        'validate_loss': validate_loss, 
        'test_loss': test_loss,

        'overfit': loss_alpha * overfit,
        'decline': loss_alpha * decline,
        
        'ks_overfit': loss_alpha * ks_overfit,
        'ks_decline': loss_alpha * ks_decline,
    }

    if loss_items is None:
        loss = validate_perf + loss_alpha*overfit
    elif loss_items == 'validate_auc':
        loss = (1 - metric_dict['validate']['auc']) + loss_alpha*abs(metric_dict['train']['auc'] - metric_dict['validate']['auc'])
    else:
        t_print('loss_items =', loss_items)
        loss = np.mean([loss_dict[k] for k in loss_items])

    if reverse and reversed_as_inf:
        loss = float('inf')


    if auc_overfit*2 < ks_overfit :
        loss = float('inf')
    
    if (overfit_rate is not None) and (metric_dict['train']['auc'] > .5) and (metric_dict['validate']['auc'] > .5):
        auc_overfit_rate = (metric_dict['train']['auc'] - .5) / (metric_dict['validate']['auc'] - .5)
        ks_overfit_rate = metric_dict['train']['ks'] / metric_dict['validate']['ks']
        t_print(f'auc_overfit_rate = {auc_overfit_rate}, ks_overfit_rate = {ks_overfit_rate}')
        if auc_overfit_rate > overfit_rate or ks_overfit_rate > overfit_rate:
            loss = max([auc_overfit_rate, ks_overfit_rate])

    if bins_diff > 0:
        loss = bins_diff + 5

    t_print('loss =', loss)

    return loss


def early_stop(trials, eps = 1e-6, n_tail_trials = 10, n_least_trails = 20, *args):
    flag = False
    n_trials = len(trials.trials)
    print(f'n_trials = {n_trials} completed.\n')
    n = n_tail_trials # 10
    if n_trials > n_tail_trials:
        losses = [trial['result']['loss'] for trial in trials.trials]
        losses = pd.Series(losses)
        losses_head = losses.iloc[:-n]
        losses_tail = losses.iloc[-n:]
        # losses_range = losses.max() - losses.min()
        t_print('losses_tail.describe() =', losses_tail.describe())
        # if losses.std() < m_losses_stat:
        decline_gap = losses_head.min() - losses_tail.min()
        t_print('decline_gap =', decline_gap)
        if n_trials > n_least_trails and decline_gap < eps:
            t_print(f'early_stop(): flag = True')
            flag = True
    return flag, args


def hyperopt_func_v0(X3: list, Y3: list, fillna = None, **kwargs):
    max_evals = kwargs.get('max_evals', 100)
    loss_alpha = kwargs.get('loss_alpha', 1)
    loss_items = kwargs.get('loss_items', ['validate_loss', 'overfit'])
    eps = kwargs.get('eps', .001)
    n_tail_trials = kwargs.get('n_tail_trials', 10)
    n_least_trails = kwargs.get('n_least_trails', 20)
    reversed_as_inf = kwargs.get('reversed_as_inf', True)
    min_bins = kwargs.get('min_bins', 0)
    overfit_rate = kwargs.get('overfit_rate', None)
    space = kwargs.get('space', 'hyperopt_func_v0_space')
    space = hyperopt_func_space_dict[space]

    objective_func = functools.partial(
        objective_func_with_params, 
        XY3 = [X3, Y3],
        loss_func=functools.partial(
            loss_func_alpha, 
            loss_alpha = loss_alpha, 
            loss_items=loss_items,
            reversed_as_inf = reversed_as_inf,
            min_bins = min_bins,
            overfit_rate = overfit_rate,
            ),
        fillna = fillna
        )
    trials = Trials()
    best_params = fmin(fn=objective_func,
                space=space,
                algo=tpe.suggest,
                max_evals=max_evals,
                trials=trials,
                early_stop_fn=functools.partial(
                    early_stop, 
                    eps = eps, 
                    n_least_trails = n_least_trails, 
                    n_tail_trials = n_tail_trials, 
                    ),
                rstate=np.random.default_rng(0),
                )
    best_params = space_eval(space, best_params)
    if True:
        best_params['scale_pos_weight'] = sum(Y3[0] == 0) / sum(Y3[0] == 1)
    t_print(f're-train model at params =', best_params)
    best_model = XGBClassifier(**best_params, random_state=0)
    best_model.fit(X3[0], Y3[0])
    return best_model


def modeling(X_train: pd.DataFrame, y_train: pd.Series, params = {}, clf_tag = 'GB', fillna = {}, weight = None, XY3 = None):

    # X_train, fillna = x_fillna(X_train, clf_tag=clf_tag, fillna=fillna)
    t_print('to_numpy()')
    t_print(f'X_train\n{X_train}')
    X_train = X_train.to_numpy()
    t_print(f'y_train\n{y_train}')
    y_train = y_train.to_numpy()
    if weight is not None:
        if isinstance(weight, pd.Series):
            weight = weight.to_numpy()

    t0 = time()
    if 'hyperopt' not in clf_tag:
        clf = get_clf(clf_tag=clf_tag, params=params)
        t_print('clf.fit(X_train, y_train)')
        clf.fit(X_train, y_train, weight)

    else:
        X3, Y3 = XY3
        t_print('clf = hyperopt_func_v0(X3=X3, Y3=Y3, **params)')
        clf = hyperopt_func_v0(X3=X3, Y3=Y3, **params)

    t1 = time()

    fit_time = t1 - t0
    if fit_time < 1:
        sleep(1 - fit_time)

    t_print(f'fit_time = {fit_time:.1f}s')
    clf_str = clf.__str__().replace('\n', '').replace(' ', '').replace(',', ', ')
    t_print(clf_str)

    clf_json = {}
    clf_json['clf_str'] = clf_str
    clf_json['clf_tag'] = clf_tag
    clf_json['clf_params'] = dict_to_json(params)
    clf_json['fit_time'] = fit_time
    clf_json['get_params'] = dict_to_json(clf.get_params(deep=False))

    return clf, clf_json, fillna


def get_auc_ks(y, y_proba, sample_weight = None, dtype = 'dict'):
    r = {}
    t0 = time()
    y, y_proba = to_numpy([y, y_proba])
    r['auc'] = roc_auc_score(y, y_proba, sample_weight=sample_weight)
    r['ks'] = ks_score(y, y_proba, sample_weight=sample_weight)
    if dtype == 'pd.Series':
        r = pd.Series(r)
    t1 = time()
    delta_t = t1 - t0
    if delta_t > 60:
        t_print(f'warning: get_auc_ks() takes too long, T = {delta_t:1.0f}s')
    return r


def get_stat_metric(clf, X, y, sample_weight = None):
    if sample_weight is not None and len(sample_weight) == 0:
        sample_weight = None
    data_json = {}
    data_json['stat'] = {}
    data_json['stat']['x'] = {}
    data_json['stat']['y'] = {}
    data_json['stat']['x']['shape'] = list(X.shape)
    data_json['stat']['y']['shape'] = list(y.shape)
    if len(X) == len(y) > 0:
        y_proba = clf.predict_proba(X.to_numpy())[:, 1]
        data_json['stat']['y']['count'] = y.value_counts().to_dict()
        data_json['stat']['y']['percent'] = (y.value_counts()/len(y)).to_dict()
        data_json['metric'] = get_auc_ks(y, y_proba, sample_weight=sample_weight)
    else:
        data_json['stat']['y']['count'] = {'0': 0, "1": 0}
        data_json['stat']['y']['percent'] = {'0': float('nan'), "1": float('nan')}
        data_json['metric'] = {'auc': float('nan'), "ks": float('nan')}
    return data_json


def modeling_tri(X, y, settings = {}, fillna = {}, params = {}, mode = None, weight = None, more_func = lambda *s: {}):

    clf_tag = settings['clf_tag']

    ds = settings['DS.random_state']
    test_size = settings['DS.test_size']
    X_train, X_valid, X_test, y_train, y_valid, y_test = get_xy_tri(X, y, ds, test_size = test_size)
    del X
    del y

    if 'selected_dim' in settings:
        t_print('warning: selected_dim in setting')
        X_train = X_train.loc[:, settings['selected_dim']]
        X_valid = X_valid.loc[:, settings['selected_dim']]
        X_test = X_test.loc[:, settings['selected_dim']]

    if weight is not None:
        train_weight, validate_weight, test_weight = weight
    else:
        train_weight = validate_weight = test_weight = None

    if X_train.columns.to_list() == X_valid.columns.to_list() == X_test.columns.to_list():
        keys = X_train.columns.to_list()

    X3 = [X_train, X_valid, X_test]
    Y3 = [y_train, y_valid, y_test]

    clf, clf_json, fillna = modeling(X_train, y_train, params=params, clf_tag=clf_tag, fillna=fillna, weight=train_weight, XY3=[X3, Y3])

    all_json = {
        'main': settings,
        'clf': clf_json,
        'data_set': {
            'train': get_stat_metric(clf, X_train, y_train, sample_weight = train_weight),
            'validate': get_stat_metric(clf, X_valid, y_valid, sample_weight = validate_weight),
            'test': get_stat_metric(clf, X_test, y_test, sample_weight = test_weight),
        },
        'more': more_func(clf, X3, Y3)
    }

    t_print(f"clf_params:\n{json_dumps(all_json['clf']['clf_params'])}")
    t_print(f"train:\n{json_dumps(all_json['data_set']['train']['metric'])}")
    t_print(f"validate:\n{json_dumps(all_json['data_set']['validate']['metric'])}")
    t_print(f"test:\n{json_dumps(all_json['data_set']['test']['metric'])}")

    return clf, all_json, fillna, keys


def get_clf_path(path = './', verbose = False, n = 0):
    s = os.listdir(path)
    if 'clf.json' in s:
        r = path
        if verbose:
            t_print(f'save_path = {path}')
            json_path = os.path.join(path, 'clf.json')
            clf_json = json_load(json_path)
            if 'main'in clf_json:
                show_json = {k: v for k, v in clf_json['main'].items() if k in ['prefix', 'exp_name']}
                t_print(show_json)
    else:
        path = ls_dirs(path)[n]
        r = get_clf_path(path, verbose = verbose, n = n)
    return r


def get_keys_and_clf(path, verbose = False, n = 0, pmml = False, clf_file = None, fillna=False):
    
    path = get_clf_path(path, verbose = verbose, n = n)
    clf_files = os.listdir(path)

    if 'keys.json' in clf_files:
        keys = json_load(os.path.join(path, 'keys.json'))
    elif 'keys_list_pickle' in clf_files:
        keys = pickle_load(os.path.join(path, 'keys_list_pickle'))
    else:
        t_print('ERRROR: could not find keys.')

    if clf_file in clf_files:
        clf = pickle_load(os.path.join(path, clf_file))
    elif 'clf_pickle' in clf_files:
        clf = pickle_load(os.path.join(path, 'clf_pickle'))
    elif 'clf_sklearn_pickle' in clf_files:
        clf = pickle_load(os.path.join(path, 'clf_sklearn_pickle'))
    else:
        t_print('ERRROR: could not find clf.')

    if isinstance(clf, GradientBoostingClassifier) and '_loss' not in dir(clf):
        print('warning: low version model file and fix _loss.')
        clf._loss = clf.__dict__['loss_']

    if pmml:
        if 'clf.pmml' in clf_files and 'vars_map_main.json' in clf_files:
            clf_pmml_path = os.path.join(path, 'clf.pmml')
            clf = Model.fromFile(clf_pmml_path)
            vars_map_main_path = os.path.join(path, 'vars_map_main.json')
            vars_map_main = json_load(vars_map_main_path)
            keys = [k[0] for k in sorted(vars_map_main.items(), key=lambda a: int(a[1][1:]))]
        else:
            t_print('')

    if fillna:
        if 'fillna.json' in clf_files:
            fillna_json = json_load(os.path.join(path, 'fillna.json'))
            r = keys, clf, fillna_json
        else:
            t_print('ERRROR: could not find fillna.json')
    else:
        r = keys, clf

    return r


def get_FI(path, verbose = False, n = 0, normalized = True):
    keys, clf = get_keys_and_clf(path, verbose = verbose, n = n)
    if 'feature_importances_' in dir(clf):
        if isinstance(clf, LGBMClassifier) or isinstance(clf, XGBClassifier):
            clf.importance_type = 'gain'
            if clf.importance_type == 'gain':
                # t_print('info: set clf.importance_type = gain')
                pass
            else:
                raise ValueError
        FI_s = pd.Series(dict(zip(keys, clf.feature_importances_)), dtype=float)
    else:
        t_print(f"warning: 'feature_importances_' not in {clf.__str__()}")
        FI_s = pd.Series({k: float('nan') for k in keys})
    if normalized:
        FI_s = FI_s / FI_s.sum()
    return FI_s


def save_FI_csv(clf_path: str, exist='skip') -> pd.DataFrame:
    clf_path = get_clf_path(clf_path)
    FI_sr = get_FI(clf_path)
    FI_df = pd.DataFrame({'importances': FI_sr})
    FI_df.index.name = 'feature'
    FI_df.reset_index(inplace=True)
    FI_df.index.name = 'index'
    FI_df.index += 1
    FI_csv_path = os.path.join(clf_path, 'FI.csv')
    if os.path.exists(FI_csv_path) and exist == 'skip':
        print('FI.csv exists.')
    else:
        FI_df.to_csv(FI_csv_path)
    return FI_df


def fix_pmml_timestamp(path, timestamp):
    if isinstance(timestamp, str) and len(timestamp) == 17:
        timestamp = datetime.datetime.strptime(timestamp, '%Y%m%d%a%H%M%S')
    timestamp_str = timestamp.strftime('%Y-%m-%dT%H:%M:%S+08:00')
    pmml_str = read_txt(path)
    Timestamp_pattern = r'<Timestamp>.+</Timestamp>'
    pmml_str = re.sub(Timestamp_pattern, f'<Timestamp>{timestamp_str}</Timestamp>', pmml_str)
    save_txt(pmml_str, path)
    return None


def dump_clf_pmml(clf, path, with_repr=True, timestamp = None):
    pipeline = PMMLPipeline([
        ("classifier", clf)
    ])
    sklearn2pmml(pipeline, path, with_repr=with_repr)
    if timestamp is not None:
        fix_pmml_timestamp(path, timestamp)
    return None


def save_clf_pmml(clf_path: str, exist='skip', timestamp = None) -> None:
    clf_path = get_clf_path(clf_path)
    keys, clf = get_keys_and_clf(clf_path)
    vars_map_all = {key: f'x{k+1}' for k, key in enumerate(keys)}
    clf_pmml_path = os.path.join(clf_path, 'clf.pmml')
    if os.path.exists(clf_pmml_path) and exist == 'skip':
        print('clf.pmml exists.')
    else:
        dump_clf_pmml(clf, clf_pmml_path, timestamp = timestamp)
    vars_map_all_path = os.path.join(clf_path, 'vars_map_all.json')
    if os.path.exists(vars_map_all_path) and exist == 'skip':
        print('vars_map_all.json exists.')
    else:
        json_dump(vars_map_all, vars_map_all_path)
    return None


def save_pmml_keys(clf_path, exist='skip') -> None:
    clf_path = get_clf_path(clf_path)
    clf_pmml_path = os.path.join(clf_path, 'clf.pmml')
    model = Model.fromFile(clf_pmml_path)
    main_vars = model.inputNames
    vars_map_all_path = os.path.join(clf_path, 'vars_map_all.json')
    vars_map_all = json_load(vars_map_all_path)
    vars_map_main = {k: v for k, v in vars_map_all.items() if v in main_vars}
    vars_map_main_path = os.path.join(clf_path, 'vars_map_main.json')
    if os.path.exists(vars_map_main_path) and exist == 'skip':
        print('vars_map_main.json exists.')
    else:
        json_dump(vars_map_main, vars_map_main_path)
    return None


def dump_files_hash(path, dump = True, timestamp = None):
    t_print('dir_path = ', path)
    files = sorted(os.listdir(path))
    file_paths = [os.path.join(path, f) for f in files if os.path.isfile(os.path.join(path, f))]
    file_hash_dict = {}
    for k, file_path in enumerate(file_paths):
        f = os.path.basename(file_path)
        t_print(f'{k+1}/{len(file_paths)}', f)
        file_hash_dict[f] = {}
        file_hash_dict[f]['size'] = os.path.getsize(file_path)
        file_hash_dict[f]['sha256'] = get_file_hash(file_path, alg='sha256')
        file_hash_dict[f]['md5'] = get_file_hash(file_path, alg='md5')
    if dump:
        if timestamp is None:
            timestamp = get_timestamp()
        json_name = f'file_hash_{timestamp}.json'
        json_path = os.path.join(path, json_name)
        json_dump(file_hash_dict, json_path)
    t_print('end of dump_files_hash()')
    return file_hash_dict


def add_mdl_files(path, hash_files = False, timestamp = None):

    clf_path = get_clf_path(path)

    try:
        save_FI_csv(clf_path)
        t_print('FI.csv saved.')
    except:
        print('ERROR: save_FI_csv(path)')

    try:
        save_clf_pmml(clf_path, timestamp = timestamp)
        t_print('clf.pmml saved.')
    except Exception as e:
        print(f'ERROR: save_clf_pmml(): {e}')

    try:
        save_pmml_keys(clf_path)
        t_print('clf key files saved.')
    except:
        print('ERROR: save_pmml_keys(path)')

    if hash_files:
        dump_files_hash(clf_path)
    
    return None


def save_result_files(clf, all_json, fillna, keys, add_pmml = True):

    all_json = dict_to_json(all_json)

    try:
        if isinstance(clf, AutoML):
            clf = clf.model.estimator
    except:
        t_print('warning: No AutoML')
        pass

    results_path = all_json['main']['results_path']
    prefix = all_json['main']['prefix']
    exp_name = all_json['main']['exp_name']
    save_name_0 = all_json['main']['save_name_0']

    exp_path = os.path.join(results_path, prefix, exp_name)

    save_results = True

    if os.path.exists(exp_path):
        exps = [k[:-17] for k in os.listdir(exp_path)]
        if save_name_0 in exps:
            save_results = False
            save_name_0_path = os.path.join(exp_path, save_name_0)
            save_name_0_path = replace_slash(save_name_0_path)
            t_print(f'{save_name_0_path}* exists')

    if save_results:
        timestamp = get_timestamp()
        save_name = save_name_0 + timestamp
        save_path = os.path.join(exp_path, save_name)
        save_path = replace_slash(save_path)
        all_json['main']['timestamp'] = timestamp
        all_json['main']['uuid_hex'] = get_uuid_hex()
        all_json['main']['save_name'] = save_name
        all_json['main']['save_path'] = save_path
        os.makedirs(save_path)

        clf_path = os.path.join(save_path, 'clf_pickle')
        pickle_dump(clf, clf_path)

        clf_json_path = os.path.join(save_path, 'clf.json')
        json_dump(all_json, clf_json_path)

        fillna = {k: v for k, v in fillna.items() if k in keys}
        fillna_json_path = os.path.join(save_path, 'fillna.json')
        json_dump(fillna, fillna_json_path)

        keys_json_path = os.path.join(save_path, 'keys.json')
        json_dump(keys, keys_json_path)

        if add_pmml :
            add_mdl_files(save_path, timestamp=timestamp)

        dump_files_hash(save_path, timestamp = timestamp)

        t_print(f'{save_path} saved')

    return None


def nan_addition(a, b):
    if pd.isna(a):
        r = b
    elif pd.isna(b):
        r = a
    else:
        r = a + b
    return r


def compress_and_encode_str(s, tag = False):
    b = s.encode('utf-8')
    compressed = gzip.compress(b)
    encoded = base64.b64encode(compressed).decode('utf-8')
    if tag:
        encoded = f'<gzip_base64>{encoded}</gzip_base64>'
    return encoded


def parse_clf_json(path: str):
    d = json_load(path)
    row = pd.Series(dtype=object)
    row['prefix'] = d['main']['prefix']
    row['exp_name'] = d['main']['exp_name']
    row['save_name'] = d['main']['save_name']
    row['uuid_hex'] = d['main'].get('uuid_hex', get_hash(row['save_name'], alg='md5'))
    row['exp_path'] = os.path.join(
        d['main']['results_path'], 
        row['prefix'],
        row['exp_name']
        ).replace('\\', '/')
    # row['save_path'] = d['main']['save_path']
    row['save_path'] = os.path.dirname(path).replace('\\', '/')
    row['date'] = datetime.datetime.strptime(d['main']['save_name'][-17:], "%Y%m%d%a%H%M%S")
    row['clf_tag'] = d['main']['clf_tag']
    row['clf_str'] = d['clf']['clf_str']
    row['clf_params'] = d['clf']['clf_params']
    row['fit_time'] = round(d['clf']['fit_time'], 1)
    row['len'] = sum([d['data_set'][k]['stat']['x']['shape'][0] for k in ['train', 'validate', 'test']])
    row['dim'] = d['data_set']['train']['stat']['x']['shape'][1]

    row['train_auc'] = d['data_set']['train']['metric']['auc']
    row['validate_auc'] = d['data_set']['validate']['metric']['auc']
    row['test_auc'] = d['data_set']['test']['metric']['auc']
    row['train_ks'] = d['data_set']['train']['metric']['ks']
    row['validate_ks'] = d['data_set']['validate']['metric']['ks']
    row['test_ks'] = d['data_set']['test']['metric']['ks']

    row['abs_overfit_auc'] = row['train_auc'] - row['validate_auc']
    row['rel_overfit_auc'] = row['abs_overfit_auc'] / (row['validate_auc'] - .5) if (row['validate_auc'] - .5) > 0 else float('nan')

    row['abs_decline_auc'] = row['validate_auc'] - row['test_auc']
    row['rel_decline_auc'] = row['abs_decline_auc'] / (row['validate_auc'] - .5) if (row['validate_auc'] - .5) > 0 else float('nan')

    row['abs_auc_overfit_and_decline'] = nan_addition(abs(row['abs_overfit_auc']), abs(row['abs_decline_auc']))

    row['overfit'] = row['rel_overfit_auc']
    row['decline'] = row['rel_decline_auc']

    # row['overfit'] = (row['train_ks'] - row['validate_ks']) / row['validate_ks'] if row['validate_ks'] != 0 else float('nan')
    # row['decline'] = (row['validate_ks'] - row['test_ks'])  / row['validate_ks'] if row['validate_ks'] != 0 else float('nan')
    # row['abs_overfit'] = row['train_ks'] - row['validate_ks']  
    # row['abs_decline'] = row['validate_ks'] - row['test_ks'] 
    # row['abs_overfit_and_decline'] = abs(row['abs_overfit']) + abs(row['abs_decline'])
    
    if 'selected_dim' in d['main']:
        row['selected_dim'] = json.dumps(d['main']['selected_dim'], ensure_ascii=False)
    else:
        row['selected_dim'] = float('nan')

    if 'more' in d:
        row['more'] = compress_and_encode_str(
            json.dumps(
                dict_to_json(d['more'], 6), 
                ensure_ascii=False
                ), 
                tag = True
                )
    else:
        row['more'] = '{}'

    return row


def show_mdl_stat_info(stat_df, abstract_cols = ['len', 'dim', 'validate_ks', 'test_ks', 'overfit', 'decline'], verbose = True):
    # abstract_cols = ['dim', 'validate_ks', 'test_ks', 'overfit', 'decline']
    n = len(stat_df)
    min_s = stat_df.loc[:, abstract_cols].astype(float).min().round(3)
    mean_s = stat_df.loc[:, abstract_cols].astype(float).mean().round(3)
    max_s = stat_df.loc[:, abstract_cols].astype(float).max().round(3)
    info_df = pd.DataFrame({
        'min': min_s, 
        'mean': mean_s, 
        'max': max_s})
    info_df = info_df.T
    info_df['len'] = info_df['len'].astype(int)
    info_df['dim'] = info_df['dim'].astype(int)
    if verbose:
        prefix = stat_df['prefix'].unique().tolist()
        print(f'# prefix = {prefix}, n = {n}', )
        print(info_df)
    return info_df


def parse_clfs(path: str, save = True):
    paths = ls_dirs_r(path, any_ls='^clf.json$')
    stat_dict = {k: os.path.join(path,'clf.json') for k, path in enumerate(paths, start=1)}
    stat_dict = multiprocessing_apply(parse_clf_json, stat_dict)
    df = pd.DataFrame(stat_dict)
    df = df.transpose()

    df['ks_std'] = df.loc[:, ['train_ks', 'validate_ks', 'test_ks']].std(axis=1)
    df.sort_values(by=['exp_name', 'validate_auc'], ascending=[True, False], inplace=True)

    # df['choose'] = False
    # choose_df = df.loc[:, ['validate_auc', 'test_auc', 'validate_ks', 'test_ks', 'overfit', 'decline']]
    # choose_df['validate_auc'] = (choose_df['validate_auc']).apply(round_nan)
    # choose_df['test_auc'] = (choose_df['test_auc']).apply(round_nan)
    # choose_df['validate_ks'] = (choose_df['validate_ks']).apply(round_nan)
    # choose_df['test_ks'] = (choose_df['test_ks']).apply(round_nan)
    # choose_df['overfit'] = (-choose_df['overfit'].abs()).apply(round_nan)
    # choose_df['decline'] = (-choose_df['decline']).apply(round_nan)
    # choose_idx = choose_one_idx(choose_df)
    # df.loc[choose_idx, 'choose'] = True

    # sort_df = df.loc[:, ['validate_ks', 'test_ks']]#.applymap(lambda a: round(a, 3))
    # sort_df.loc[df['overfit'] >= 1, :] = float('nan')
    # sort_df.sort_values(by=['validate_ks', 'test_ks'], ascending = False, inplace=True)
    # df = df.loc[sort_df.index, :]
    # df.loc[df.index.to_list()[0], 'choose'] = True

    df.index = pd.Index(range(1, len(df) + 1), name='index') 
    if save:
        if type(save) != str:
            save_path = path + '-stat.csv'
        else:
            save_path = save
        df.to_csv(save_path)
        t_print(f'{save_path} saved. len_clfs = {len(df)}')
    show_mdl_stat_info(df)
    return df


def auto_params_v2(X, y, settings_dict, weight=None, more_func = more_func, test_only = False, add_pmml = True, fillna = None):

    if type(X) == tuple:
        X = list(X)
    
    if type(y) == tuple:
        y = list(y)

    X = [Xk.copy().astype(float) for Xk in X]
    y = [yk.copy() for yk in y]
    
    if not (len(X) == len(y) == 3):
        raise ValueError('X and y must be list of length 3')

    if isinstance(fillna, int) or isinstance(fillna, float) :
        fillna = {k: fillna for k in X[0].keys()}

    if isinstance(fillna, str) and fillna == 'mean':
        fillna = X[0].mean()

    if isinstance(fillna, pd.Series):
        fillna = fillna.to_dict()
        print('fillna =', fillna)

    t_print('fillna')
    [Xk.fillna(fillna, inplace = True) for Xk in X]

    if not isinstance(fillna, dict):
        raise ValueError('fillna must be dict or pd.Series')

    settings_dict_list = get_params_dict_list(settings_dict)
    shuffle(settings_dict_list)

    global exp_name_global
    global iter_global

    eta = 0
    t0 = time()
    for k, settings in enumerate(settings_dict_list):
        t1 = time()
        cnt = k + 1
        t_print(f'{cnt}/{len(settings_dict_list)}')
        settings, params = fix_settings(settings, X)
        exp_name_global = settings['exp_name']
        
        if run_clf(settings):
            t_print(settings)
            t_print(params)
            t_print(f'T_one = 0 at {date_str()}')
            if not test_only:
                sleep(1)
                clf, all_json, _, keys = modeling_tri(X, y, settings=settings, params=params, weight=weight, more_func=more_func)
                save_result_files(clf, all_json, fillna, keys, add_pmml=add_pmml)
            else:
                t_print('test_only = True')
            t2 = time()
            eta = (t2 - t0) / (k+1) * (len(settings_dict_list) - cnt)
            t_print(f'T_one = {t2 - t1:1.0f}s')
            iter_global = 0
            exp_name_global = ''
        t_print(f'All will end at {date_str(eta)}.\n#')
        
    stat_df = parse_clfs(os.path.join(settings_dict['results_path'], settings['prefix']))
    t_print(f'auto_params_v2() ends at {date_str()}.')
    t_print(f'end of', settings['prefix'])
    print('\n'*5)
    return stat_df


def get_FI_top_n(mean_FI_sr, th = .999):
    if type(mean_FI_sr) == dict:
        mean_FI_sr = pd.Series(mean_FI_sr)
    mean_FI_sr = mean_FI_sr[mean_FI_sr > 0]
    if th == 1:
        keys = mean_FI_sr.index.to_list()
    elif th < 1:
        for k in range(1, len(mean_FI_sr)+1):
            FI_nlargest = mean_FI_sr.nlargest(k)
            if FI_nlargest.sum() >= th:
                break
        keys = FI_nlargest.keys().to_list()
    r = [k for k in mean_FI_sr.index if k in keys]
    return r


def get_FI_from_exp(exp_path, verbose = False, n = 0, th = .999, how = 'union'):
    if isinstance(exp_path, pd.Series):
        exp_path = exp_path.to_list()
    if how not in ['union', 'intersec']:
        raise ValueError(f'how = {how}')
    if type(exp_path) == list:
        paths = exp_path
    elif type(exp_path) == str:
        paths = ls_dirs(exp_path)
        if len(paths) == 0:
            paths = [exp_path]
    top_keys_ls = []
    k = 0
    for path in paths:
        k += 1
        FI_df = get_FI(path, verbose=verbose, n = n)
        top_keys = get_FI_top_n(FI_df, th = th)
        top_keys_ls.append(top_keys)
        t_print(f'{k}/{len(paths)} top_features: {len(FI_df)} -> {len(top_keys)}')
    merged_features = set(top_keys_ls[0])
    for features in top_keys_ls:
        if how == 'union':
            merged_features |= set(features)
        elif how == 'intersec':
            merged_features &= set(features)
    merged_features = sorted(merged_features)
    if merged_features:
        t_print('len(merged_features) =', len(merged_features))
    else:
        raise ValueError(f'merged_features = [] in {exp_path}')
    return merged_features


def check_xy_data(X = [], y = [], found_nan = 'raise'):
    for Xk in X:
        print('Xmin =', Xk.min().min())
        print('Xmax =', Xk.max().max())
        if Xk.isna().any().any() and found_nan == 'raise':
            raise ValueError('X has nan')
        elif Xk.isna().any().any():
            print('X has nan')
        else:
            print('X has no nan')
    for yk in y:
        print(yk.value_counts().to_dict())
        if yk.isna().any() :
            raise ValueError('y has nan')
    return None


def split_xy_data(y_df: pd.DataFrame, x_df: pd.DataFrame, how = 'intersec', fillna = -10, found_nan='raise'):
    
    t_print('fillna =', fillna)
    y_df = y_df.loc[~y_df['data_split'].isna(), :]
    idx = y_df.index.intersection(x_df.index)
    n1 = len(y_df)
    n2 = len(idx)
    t_print(f'coverage: {n2 / n1} (= {n2}/{n1})')
    y_df = y_df.loc[idx, :]
    y_df['y'] = y_df['y'].astype(int)
    x_df = x_df.loc[idx, :].astype(float)
    
    if isinstance(fillna, str) and fillna == 'mean':
        fillna = x_df.loc[y_df['data_split'] == 'train', :].mean()
        t_print('fillna =', fillna.to_dict())
    if fillna is not None:
        x_df = x_df.fillna(fillna)
    
    split_ls = ['train', 'validate', 'test']
    y_dfs = [y_df.loc[y_df['data_split'] == split, 'y'] for split in split_ls]
    x_dfs = [x_df.loc[y_df['data_split'] == split, :  ] for split in split_ls]
    
    t_print(f'y_dfs = {y_dfs}', )
    t_print(f'x_dfs = {x_dfs}', )
    
    check_xy_data(x_dfs, y_dfs, found_nan=found_nan)
    
    return y_dfs, x_dfs


#%%


def rewrited_stable_cumsum(arr, axis=None, rtol=1e-05, atol=1e-08):
    out = np.cumsum(arr, axis=axis, dtype=np.float64)
    return out


def rewrited_binary_clf_curve(y_true, y_score, pos_label=None, sample_weight=None):
    pos_label = 1
    if isinstance(y_true, pd.Series):
        y_true = y_true.to_numpy()
    if isinstance(y_score, pd.Series):
        y_score = y_score.to_numpy()
    y_true = y_true == pos_label
    desc_score_indices = np.argsort(y_score, kind="mergesort")[::-1]
    y_score = y_score[desc_score_indices]
    y_true = y_true[desc_score_indices]
    if sample_weight is not None:
        weight = sample_weight[desc_score_indices]
    else:
        weight = 1.0
    distinct_value_indices = np.where(np.diff(y_score))[0]
    threshold_idxs = np.r_[distinct_value_indices, y_true.size - 1]
    tps = rewrited_stable_cumsum(y_true * weight)[threshold_idxs]
    if sample_weight is not None:
        fps = rewrited_stable_cumsum((1 - y_true) * weight)[threshold_idxs]
    else:
        fps = 1 + threshold_idxs - tps
    return fps, tps, y_score[threshold_idxs]


def rewrited_roc_curve(y_true, y_score, pos_label=None, sample_weight=None, drop_intermediate=True):
    fps, tps, thresholds = rewrited_binary_clf_curve(
        y_true, y_score, pos_label=pos_label, sample_weight=sample_weight
    )
    if drop_intermediate and len(fps) > 2:
        optimal_idxs = np.where(
            np.r_[True, np.logical_or(np.diff(fps, 2), np.diff(tps, 2)), True]
        )[0]
        fps = fps[optimal_idxs]
        tps = tps[optimal_idxs]
        thresholds = thresholds[optimal_idxs]
    tps = np.r_[0, tps]
    fps = np.r_[0, fps]
    thresholds = np.r_[thresholds[0] + 1, thresholds]
    fpr = fps / fps[-1]
    tpr = tps / tps[-1]
    return fpr, tpr, thresholds


def rewrited_auc(x, y):
    direction = 1
    dx = np.diff(x)
    if np.any(dx < 0):
        if np.all(dx <= 0):
            direction = -1
    area = direction * np.trapz(y, x)
    if isinstance(area, np.memmap):
        area = area.dtype.type(area)
    return area


def rewrited_binary_roc_auc_score(y_true, y_score, sample_weight=None):
    if isinstance(y_true, pd.Series):
        y_true = y_true.to_numpy()
    if isinstance(y_score, pd.Series):
        y_score = y_score.to_numpy()
    fpr, tpr, _ = rewrited_roc_curve(y_true, y_score, sample_weight=sample_weight)
    ret = rewrited_auc(fpr, tpr)
    ret = float(ret)
    return ret

roc_curve = rewrited_roc_curve
roc_auc_score = rewrited_binary_roc_auc_score

def ks_score(y, y_proba, sample_weight = None):
    fpr, tpr = roc_curve(y, y_proba, sample_weight=sample_weight)[:2]
    r = np.abs(tpr - fpr).max()
    return r



